@@ -13,11 +13,12 @@ use crate::{
         error::QueryExecutionError,
         Address, ByteCode, Program, Result, ScopeId, ScopedSlot, Value,
     },
-    Array,
+    Array, InputError,
 };
 use itertools::Itertools;
 use std::{
     cell::{RefCell, RefMut},
+    iter::Fuse,
     rc::Rc,
 };
 use thiserror::Error;
@@ -158,13 +159,14 @@ enum OnFork {
     TryAlternative,
     CatchLabel(LabelId),
     Iterate,
+    IterateInput,
 }
 
 impl Environment {
     fn new(state: <State as Undo>::UndoToken) -> Self {
         Self {
             next_label_id: 0,
-            forks: vec![(state, OnFork::Nop)],
+            forks: vec![(state, OnFork::IterateInput)],
         }
     }
 
@@ -429,33 +431,47 @@ impl Machine {
         }
     }
 
-    pub fn run(&mut self, value: Value) -> ResultIterator {
+    pub fn start<I: Iterator<Item = Result<Value, InputError>>>(
+        &mut self,
+        input: I,
+    ) -> ResultIterator<I> {
         let mut state = State::new(self.program.entry_point);
-        state.push(value);
         let env = Environment::new(state.save());
         ResultIterator {
             program: self.program.clone(),
             env,
             state,
+            input: input.fuse(),
         }
     }
 }
 
-pub struct ResultIterator {
+pub struct ResultIterator<I: Iterator<Item = Result<Value, InputError>>> {
     program: Rc<Program>,
     env: Environment,
     state: State,
+    input: Fuse<I>,
 }
 
-impl Iterator for ResultIterator {
+impl<I: Iterator<Item = Result<Value, InputError>>> Iterator for ResultIterator<I> {
     type Item = Result<Value>;
 
     fn next(&mut self) -> Option<Self::Item> {
-        run_code(&self.program, &mut self.state, &mut self.env)
+        run_code(
+            &self.program,
+            &mut self.state,
+            &mut self.env,
+            &mut self.input,
+        )
     }
 }
 
-fn run_code(program: &Program, state: &mut State, env: &mut Environment) -> Option<Result<Value>> {
+fn run_code(
+    program: &Program,
+    state: &mut State,
+    env: &mut Environment,
+    input: &mut impl Iterator<Item = Result<Value, InputError>>,
+) -> Option<Result<Value>> {
     let mut err: Option<QueryExecutionError> = None;
     log::trace!("Start from environment {:?}", env);
     'backtrack: loop {
@@ -539,6 +555,26 @@ fn run_code(program: &Program, state: &mut State, env: &mut Environment) -> Opti
                         }
                     }
                 }
+                (OnFork::IterateInput, None) => {
+                    match input.next() {
+                        None => return None,
+                        Some(Err(e)) => {
+                            // Don't push fork again so we don't infinitely see input errors
+                            // even when the input stream gave us infinite errors.
+                            return Some(Err(e.into()));
+                        }
+                        Some(Ok(v)) => {
+                            state.undo(token);
+                            env.push_fork(state, OnFork::IterateInput, state.pc);
+                            state.push(v);
+                            break 'select_fork state.save();
+                        }
+                    }
+                }
+                (OnFork::IterateInput, Some(_e)) => {
+                    env.push_fork(state, OnFork::IterateInput, state.pc);
+                    return Some(Err(err.take().unwrap()));
+                }
                 (_, Some(_)) => continue 'select_fork,
             }
             break 'select_fork token;
